@@ -1,17 +1,19 @@
 import asyncio
-import functools
 import inspect
 import logging
+import functools
 from copy import copy
-from datetime import datetime
 from enum import Enum
 from uuid import uuid4
+from datetime import datetime, timezone
 
 from tukio.dag import DAG
-from tukio.task import TaskTemplate, TaskRegistry, UnknownTaskName, TukioTask
-from tukio.utils import FutureState, Listen, SkipTask
+from tukio.utils import FutureState, Listen, SkipTask, TimeoutHandle
 from tukio.broker import get_broker, workflow_exec_topics
 from tukio.event import Event, EventSource
+from tukio.task import (
+    TaskTemplate, TaskRegistry, UnknownTaskName, TukioTask, TukioTaskError,
+)
 
 
 log = logging.getLogger(__name__)
@@ -57,12 +59,16 @@ class RescueError(WorkflowError):
 
 class WorkflowExecState(Enum):
 
-    begin = 'workflow-begin'
-    end = 'workflow-end'
-    error = 'workflow-error'
-    progress = 'workflow-progress'
-    suspend = 'workflow-suspend'
-    resume = 'workflow-resume'
+    BEGIN = 'workflow-begin'
+    END = 'workflow-end'
+    ERROR = 'workflow-error'
+    PROGRESS = 'workflow-progress'
+    SUSPEND = 'workflow-suspend'
+    RESUME = 'workflow-resume'
+
+    @classmethod
+    def values(cls):
+        return [member.value for member in cls]
 
 
 class OverrunPolicy(Enum):
@@ -77,13 +83,13 @@ class OverrunPolicy(Enum):
     """
 
     # Skip until all running instances are finished
-    skip = 'skip'
+    SKIP = 'skip'
     # Start a new workflow instance whatever the running instances
-    start_new = 'start-new'
+    START_NEW = 'start-new'
     # Skip until all running instances had been unlocked
-    skip_until_unlock = 'skip-until-unlock'
+    SKIP_UNTIL_UNLOCK = 'skip-until-unlock'
     # Abort all running instances before creating a new once
-    abort_running = 'abort-running'
+    ABORT_RUNNING = 'abort-running'
 
     @classmethod
     def get_default_policy(cls):
@@ -109,6 +115,8 @@ class OverrunPolicyHandler:
     to the list of running workflow instances (with the same template).
     """
 
+    __slots__ = ('_loop', 'template', 'policy')
+
     def __init__(self, template, loop=None):
         self._loop = loop
         # Store the workflow template for further use in policy handlers.
@@ -116,13 +124,11 @@ class OverrunPolicyHandler:
         self.policy = template.policy
 
     def new_workflow(self, running=None):
-        method = getattr(self, '_' + self.policy.name)
-        return method(running)
+        method = getattr(self, '_' + self.policy.name.lower())
+        return method(running or [])
 
     def _check_wflow(self, wflow):
-        if wflow.template.uid != self.template.uid:
-            err = 'expected template ID {}'', got {}'
-            raise ValueError(err.format(self.template.uid, wflow.template.uid))
+        return wflow.template.uid == self.template.uid
 
     def _new_wflow(self):
         return Workflow(self.template, loop=self._loop)
@@ -132,12 +138,10 @@ class OverrunPolicyHandler:
         Run a new instance of workflow only if there's no instance already
         running with the same template ID.
         """
-        if running:
-            for wflow in running:
-                self._check_wflow(wflow)
-            return None
-        else:
-            return self._new_wflow()
+        for instance in running:
+            if self._check_wflow(instance):
+                return None
+        return self._new_wflow()
 
     def _start_new(self, _):
         """
@@ -151,26 +155,20 @@ class OverrunPolicyHandler:
         running have been unlocked. Refer to the `Workflow` docstring for more
         details about locked/unlocked workflows.
         """
-        if running:
-            for wflow in running:
-                self._check_wflow(wflow)
-                if wflow.lock.locked():
-                    break
-            else:
-                return self._new_wflow()
-            # There's at least 1 locked workflow instance
-            return None
-        else:
-            return self._new_wflow()
+        for instance in running:
+            if self._check_wflow(instance):
+                if instance.lock.locked():
+                    # There's at least 1 locked workflow instance
+                    return None
+        return self._new_wflow()
 
     def _abort_running(self, running):
         """
         Abort all running instances of the workflow before creating a new one.
         """
-        if running:
-            for wflow in running:
-                self._check_wflow(wflow)
-                wflow.cancel()
+        for instance in running:
+            if self._check_wflow(instance):
+                instance.cancel()
         return self._new_wflow()
 
 
@@ -183,11 +181,16 @@ class WorkflowTemplate:
     It provides an API to easily build and update a consistent workflow.
     """
 
-    def __init__(self, uid=None, policy=None, topics=None):
+    __slots__ = ('uid', 'topics', 'policy', 'dag', 'timeout', 'schema')
+
+    def __init__(self, uid=None, policy=None, topics=None, timeout=None,
+                 schema=None):
         self.uid = uid or str(uuid4())
         self.topics = topics
         self.policy = OverrunPolicy.get(policy)
         self.dag = DAG()
+        self.timeout = timeout
+        self.schema = schema or 1
 
     @property
     def tasks(self):
@@ -249,6 +252,8 @@ class WorkflowTemplate:
                 "id": <workflow-uid>,
                 "topics": [<a-topic>, <another-topic>],
                 "policy": <policy>,
+                "timeout": <timeout>,
+                "schema": <int>,
                 "tasks": [
                     {"id": <task-uid>, "name": <name>, "config": <cfg-dict>},
                     ...
@@ -276,7 +281,9 @@ class WorkflowTemplate:
         wf_tmpl = cls(
             uid=wf_dict.get('id'),
             policy=wf_dict.get('policy'),
-            topics=wf_dict.get('topics')
+            topics=wf_dict.get('topics'),
+            timeout=wf_dict.get('timeout'),
+            schema=wf_dict.get('schema'),
         )
 
         # Tasks
@@ -287,12 +294,15 @@ class WorkflowTemplate:
             task_ids[task_tmpl.uid] = task_tmpl
 
         # Graph
+        for up_id, down_ids_set in wf_dict.get('graph', {}).items():
+            up_tmpl = task_ids[up_id]
+            for down_id in down_ids_set:
+                down_tmpl = task_ids[down_id]
+                wf_tmpl.link(up_tmpl, down_tmpl)
+
+        # Graph validation
         try:
-            for up_id, down_ids_set in wf_dict.get('graph', {}).items():
-                up_tmpl = task_ids[up_id]
-                for down_id in down_ids_set:
-                    down_tmpl = task_ids[down_id]
-                    wf_tmpl.link(up_tmpl, down_tmpl)
+            wf_tmpl.dag.validate()
         except KeyError as exc:
             raise TemplateGraphError(exc.args[0]) from exc
 
@@ -307,8 +317,10 @@ class WorkflowTemplate:
             'id': self.uid,
             'policy': self.policy.value,
             'topics': self.topics,
+            'timeout': self.timeout,
+            'schema': self.schema,
             'tasks': [],
-            'graph': {}
+            'graph': {},
         }
         for task_tmpl in self.tasks:
             wf_dict['tasks'].append(task_tmpl.as_dict())
@@ -321,7 +333,7 @@ class WorkflowTemplate:
         """
         Returns a copy of the current instance of workflow template.
         """
-        wf_tmpl = WorkflowTemplate(uid=self.uid)
+        wf_tmpl = self.__class__(uid=self.uid)
         wf_tmpl.dag = self.dag.copy()
         return wf_tmpl
 
@@ -405,6 +417,12 @@ class Workflow(asyncio.Future):
     # Inspired from the implementation of `asyncio.Task`
     _current_workflows = {}
 
+    __slots__ = (
+        'uid', '_template', '_start', '_end', 'tasks', '_tasks_by_id',
+        '_updated_next_tasks', '_done_tasks', '_internal_exc', '_must_cancel',
+        'lock', '_broker', '_source', '_committed', '_timed_out',
+    )
+
     @classmethod
     def current_workflow(cls, loop=None):
         """
@@ -413,7 +431,7 @@ class Workflow(asyncio.Future):
         None is returned when called not in the context of a Workflow.
         """
         loop = loop or asyncio.get_event_loop()
-        task = asyncio.Task.current_task(loop)
+        task = asyncio.current_task(loop)
         workflow = None
         if task:
             workflow = _get_workflow_from_task(task)
@@ -441,7 +459,7 @@ class Workflow(asyncio.Future):
         self.lock = asyncio.Lock()
         # Create the workflow in the 'locked' state when its overrun policy is
         # 'skip-until-unlock'.
-        if self.policy is OverrunPolicy.skip_until_unlock:
+        if self.policy is OverrunPolicy.SKIP_UNTIL_UNLOCK:
             self.lock._locked = True
         # Work with an event broker
         self._broker = broker or get_broker(self._loop)
@@ -451,7 +469,7 @@ class Workflow(asyncio.Future):
         )
         # A 'committed' workflow is a pending workflow not suspended
         self._committed = asyncio.Event()
-        self._committed.set()
+        self._timed_out = False
 
     @property
     def template(self):
@@ -465,6 +483,10 @@ class Workflow(asyncio.Future):
     def committed(self):
         return self._committed.is_set()
 
+    @property
+    def timed_out(self):
+        return self._timed_out
+
     @_current_workflow
     def _unlock(self, _):
         """
@@ -480,7 +502,7 @@ class Workflow(asyncio.Future):
         Adds a done callback to the passed or current task so as to unlock the
         workflow when the task gets done.
         """
-        task = task or asyncio.Task.current_task()
+        task = task or asyncio.current_task()
         task.add_done_callback(self._unlock)
 
     def _register_to_broker(self, task_tmpl, task):
@@ -538,9 +560,10 @@ class Workflow(asyncio.Future):
         """
         # A workflow can be ran only once
         if self.tasks:
-            raise RuntimeError('a workflow can be run only once!')
+            raise RuntimeError('A workflow can be run only once!')
 
         # Run the root task
+        self._committed.set()
         try:
             root_tmpl = self._template.root()
         except WorkflowRootTaskError as exc:
@@ -549,12 +572,19 @@ class Workflow(asyncio.Future):
             task = None
         else:
             # Automatically wrap input data into an event object
-            self._dispatch_exec_event(WorkflowExecState.begin, copy(event))
+            if isinstance(event, dict):
+                event = Event(event)
+            self._dispatch_exec_event(WorkflowExecState.BEGIN, copy(event))
             task = self._new_task(root_tmpl, event)
-            self._start = datetime.utcnow()
+            self._start = datetime.now(timezone.utc)
             # The workflow may fail to start at once
             if not task:
                 self._try_mark_done()
+
+        # Handle the workflow's timeout, if any.
+        if self._template.timeout:
+            TimeoutHandle(self, self._template.timeout).start()
+
         return task
 
     def _new_task(self, task_tmpl, event):
@@ -573,8 +603,10 @@ class Workflow(asyncio.Future):
             self._internal_exc = exc
             self._cancel_all_tasks()
             return None
-        task._workflow = self
-        log.debug('new task created for %s', task_tmpl)
+
+        # Link the task to this workflow and send a broker BEGIN event.
+        task.setup_workflow(self, task_tmpl)
+        log.debug('New task created for %s', task_tmpl)
 
         def next_tasks(future):
             asyncio.ensure_future(self._run_next_tasks(future))
@@ -649,17 +681,40 @@ class Workflow(asyncio.Future):
             # If the task has been skipped, we just forward the previous task
             # inputs to the next one.
             result = task.inputs
-            log.warning('task %s has been skipped', task.template)
+            log.info(
+                'Task execution %s has been skipped (%s)',
+                task.uid[:8], task.template,
+            )
+        except asyncio.TimeoutError:
+            if task.timed_out is False:
+                raise
+            # Task timed out, go on with the workflow
+            result = task.outputs
+            log.info(
+                'Task execution %s has timed out (%s)',
+                task.uid[:8], task.template,
+            )
         except asyncio.CancelledError:
-            log.warning('task %s has been cancelled', task.template)
+            # Workflow cancelled
+            log.info('Task %s has been cancelled', task.template)
+            self._try_mark_done()
+            return
+        except TukioTaskError as exc:
+            log.warning(
+                'Task %s ended on error: %s', task.template, exc.task_outputs
+            )
             self._try_mark_done()
             return
         except Exception as exc:
-            log.warning('task %s ended on exception', task.template)
+            log.warning('Task %s ended on exception', task.template)
             log.exception(exc)
             self._try_mark_done()
             return
 
+        # Ensure the result is not None (task ended without a return)
+        if result is None:
+            result = task.inputs
+
         # Ensure the workflow is not suspended, else wait for a resume.
         try:
             await self._committed.wait()
@@ -685,9 +740,18 @@ class Workflow(asyncio.Future):
                     continue
                 # Downstream task already running, join it!
                 self._join_task(next_task, event)
+                continue
+
             # Create new task
-            else:
+            try:
                 next_task = self._new_task(tmpl, event)
+            except Exception as exc:
+                # Something wrong happened during the task instantiation.
+                # We log the exception and move on since we can't decide
+                # what to do at this point.
+                log.exception(exc)
+                continue
+
             if not next_task:
                 break
 
@@ -707,10 +771,10 @@ class Workflow(asyncio.Future):
         # done callback from another task executed in the same iteration of the
         # event loop.
         if self._all_tasks_done() and not self.done():
-            exec_event = WorkflowExecState.end
+            exec_event = WorkflowExecState.END
             if self._internal_exc:
                 self.set_exception(self._internal_exc)
-                exec_event = WorkflowExecState.error
+                exec_event = WorkflowExecState.ERROR
                 data = self._internal_exc
             elif self._must_cancel:
                 super().cancel()
@@ -718,7 +782,7 @@ class Workflow(asyncio.Future):
             else:
                 self.set_result(self.tasks)
                 data = None
-            self._end = datetime.utcnow()
+            self._end = datetime.now(timezone.utc)
             self._dispatch_exec_event(exec_event, data=data)
 
     def _all_tasks_done(self):
@@ -744,7 +808,7 @@ class Workflow(asyncio.Future):
             if is_cancelled:
                 cancelled += 1
                 try:
-                    asyncio.ensure_future(task.holder.teardown())
+                    task.holder.teardown()
                 except AttributeError:
                     pass
         return cancelled
@@ -758,7 +822,7 @@ class Workflow(asyncio.Future):
         This method is intended to be called at runtime by the task itself.
         `task_tmpl_ids` must be a list (can be empty) of task template IDs.
         """
-        task = asyncio.Task.current_task(self._loop)
+        task = asyncio.current_task(self._loop)
         if task not in self.tasks:
             raise RuntimeError('task {} not executed by {}'.format(task, self))
         self._updated_next_tasks[task] = task_tmpl_ids
@@ -777,7 +841,7 @@ class Workflow(asyncio.Future):
         for task in (self.tasks - self._done_tasks):
             task.suspend()
 
-        self._dispatch_exec_event(WorkflowExecState.suspend)
+        self._dispatch_exec_event(WorkflowExecState.SUSPEND)
         log.info('workflow %s has been suspended', self)
 
     def resume(self):
@@ -797,7 +861,7 @@ class Workflow(asyncio.Future):
             event = Event(task.inputs, source=task.event_source)
             self._new_task(task.template, event)
 
-        self._dispatch_exec_event(WorkflowExecState.resume)
+        self._dispatch_exec_event(WorkflowExecState.RESUME)
         log.info('workflow %s has been resumed', self)
 
     def cancel(self):
@@ -817,6 +881,14 @@ class Workflow(asyncio.Future):
             super().cancel()
         return True
 
+    def timeout(self):
+        """
+        Called automatically after reaching the 'timeout' threshold,
+        if it exists for this workflow.
+        """
+        self.cancel()
+        self._timed_out = True
+
     def __str__(self):
         """
         Readable string representation of a workflow execution object.
@@ -846,10 +918,14 @@ class Workflow(asyncio.Future):
                 continue
             task_dict['exec'] = task.as_dict()
             # If the task is linked to a task holder, try to use its own report
-            try:
-                task_dict['exec']['reporting'] = task.holder.report()
-            except AttributeError:
-                pass
+            if hasattr(task.holder, 'report'):
+                try:
+                    task_dict['exec']['reporting'] = task.holder.report()
+                except Exception as exc:
+                    # Unexpected error from task reporting.
+                    log.error('Exception on task reporting: %s', exc)
+                    self._internal_exc = exc
+                    self._try_mark_done()
         return report
 
     def fast_forward(self, report):
@@ -928,7 +1004,7 @@ class Workflow(asyncio.Future):
                 event = Event(task_report['exec']['inputs'])
             self._new_task(task_template, event)
 
-        self._dispatch_exec_event(WorkflowExecState.begin)
+        self._dispatch_exec_event(WorkflowExecState.BEGIN)
 
 
 def new_workflow(wf_tmpl, running=None, loop=None):
